package org.eclipse.linuxtools.tmf.request;

import org.eclipse.linuxtools.tmf.event.TmfTimeRange;

/**
 * @author tgatterw
 *
 * @param <V>
 */
public class TmfFilteringDataRequest<V> extends TmfDataRequest<V>
{

    // ========================================================================
    // Attributes
    // ========================================================================

	private final Object fFilter;	// The filter object
    private final int  fNbTestedItems;   // The number of items to read and test for filtering (-1 == the whole range)

	private int[] fDataIndex;		// The index of the result events
	
    // ========================================================================
    // Constructors
    // ========================================================================

	/**
	 * @param filter	The filter object. Only events that match this filter object are returned.
	 * @param index		The start index of the request.
	 * @param offset	An offset to the time stamp of the events that are returned. 
	 * @param nbEvents			The maximum number of events that are returned, or ALL_EVENTS to return all matching events.
	 */
	public TmfFilteringDataRequest( Object filter, int index, int nbEvents )
	{
		this( filter, null, index, nbEvents, DEFAULT_BLOCK_SIZE, ALL_EVENTS );
	}

	/**
	 * @param filter	The filter object. Only events that match this filter object are returned.
	 * @param range		A range indicating start and end time of the request. 
	 * @param offset	An offset to the time stamp of the events that are returned. 
	 * @param nbEvents			The maximum number of events that are returned, or ALL_EVENTS to return all matching events.
	 */
	public TmfFilteringDataRequest( Object filter, TmfTimeRange range, int nbEvents )
	{
		this( filter, range, 0, nbEvents, DEFAULT_BLOCK_SIZE, ALL_EVENTS );
	}

	/**
	 * @param filter	The filter object. Only events that match this filter object are returned.
	 * @param range		A range indicating start and end time of the request. 
	 * @param offset	An offset to the time stamp of the events that are returned. 
	 * @param nbEvents			The maximum number of events that are returned, or ALL_EVENTS to return all matching events.
	 * @param maxBlockSize		The maximum block size that will be available in a single handleData() call. 
	 */
	public TmfFilteringDataRequest( Object filter, TmfTimeRange range, int nbEvents, int maxBlockSize )
	{
		this( filter, range, 0, nbEvents, maxBlockSize, ALL_EVENTS );
	}

	/**
	 * @param filter	The filter object. Only events that match this filter object are returned.
	 * @param range		A range indicating start and end time of the request. 
	 * @param index		The start index of the request.
	 * @param offset	An offset to the time stamp of the events that are returned. 
	 * @param nbEvents			The maximum number of events that are returned, or ALL_EVENTS to return all matching events.
	 * @param maxBlockSize		The maximum block size that will be available in a single handleData() call. 
	 */
	public TmfFilteringDataRequest( Object filter, TmfTimeRange range, int index, int nbEvents, int maxBlockSize )
	{
		this( filter, range, index, nbEvents, maxBlockSize, ALL_EVENTS );
	}

	/**
	 * @param filter	The filter object. Only events that match this filter object are returned.
	 * @param range		A range indicating start and end time of the request. 
	 * @param index		The start index of the request.
	 * @param offset	An offset to the time stamp of the events that are returned. 
	 * @param nbEvents			The maximum number of events that are returned, or ALL_EVENTS to return all matching events.
	 * @param maxBlockSize		The maximum block size that will be available in a single handleData() call. 
	 * @param nbTestedEvents	The maximum number of events that are tested, or ALL_EVENTS for no limit.
	 */
	public TmfFilteringDataRequest( Object filter, TmfTimeRange range, int index, int nbEvents, int maxBlockSize, int nbTestedEvents )
	{
		super( range, index, nbEvents, maxBlockSize );
		fFilter = filter;
		fNbTestedItems = nbTestedEvents;
	}

    // ========================================================================
    // Accessors
    // ========================================================================

	/**
	 * @return the filter object
	 */
	public Object getFilter()
	{
		return fFilter;
	}

    /**
     * @return the number of events to read and test for filtering (-1 = all)
     */
    public int getNbTestedItems() {
        return fNbTestedItems;
    }

    // ========================================================================
    // Operators
    // ========================================================================

	/**
	 * @return the index array.
	 */
	public int[] getDataIndex()
	{
		return fDataIndex;
	}

	/**
	 * @param fDataIndex the index array to set.
	 */
	public void setDataIndex( int[] fDataIndex )
	{
		this.fDataIndex = fDataIndex;
	}

}
