/*******************************************************************************
 * Copyright (c) 2009 Ericsson
 * 
 * All rights reserved. This program and the accompanying materials are
 * made available under the terms of the Eclipse Public License v1.0 which
 * accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *   Francois Chouinard - Initial API and implementation
 *   Thomas Gatterweh	- Updated experiment indexing
 *******************************************************************************/

package org.eclipse.linuxtools.tmf.trace;

import java.lang.reflect.InvocationTargetException;
import java.util.Collections;
import java.util.Comparator;
import java.util.TreeSet;
import java.util.Vector;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.linuxtools.tmf.event.TmfEvent;
import org.eclipse.linuxtools.tmf.event.TmfTimeRange;
import org.eclipse.linuxtools.tmf.event.TmfTimestamp;
import org.eclipse.linuxtools.tmf.request.TmfFilteringDataRequest;
import org.eclipse.linuxtools.tmf.signal.TmfSignalHandler;

/**
 * <b><u>TmfExperiment</u></b>
 * <p>
 * TmfExperiment presents a time-ordered, unified view of a set of TmfTraces
 * that are part of a tracing experiment. 
 * <p>
 */
public class TmfFilteredExperiment //extends TmfComponent implements ITmfRequestHandler<TmfEvent> {
extends TmfExperiment {

// TODO: Complete multi-trace experiment
// TODO: Add support for dynamic addition/removal of traces
// TODO: Add support for live streaming (notifications, incremental indexing, ...)

	// ========================================================================
    // Attributes
    // ========================================================================

    // The filter object used in this experiment
    protected final Object fFilter;
    
    // Start time
    protected final TmfTimestamp fStartTime;

    // End time
    protected final TmfTimestamp fEndTime;

    // ========================================================================
    // Constructors/Destructor
    // ========================================================================

    /**
     * @param id
     * @param traces
     */
    public TmfFilteredExperiment(String id, ITmfTrace[] traces, Object filter, TmfTimestamp startTime, TmfTimestamp endTime) {
        this(id, traces, filter, startTime, endTime, TmfTimestamp.BigBang, DEFAULT_INDEX_PAGE_SIZE, false);
    }

    /**
     * @param id
     * @param traces
     * @param waitForIndexCompletion
     */
    public TmfFilteredExperiment(String id, ITmfTrace[] traces, Object filter, TmfTimestamp startTime, TmfTimestamp endTime, boolean waitForIndexCompletion) {
        this(id, traces, filter, startTime, endTime, TmfTimestamp.BigBang, DEFAULT_INDEX_PAGE_SIZE, waitForIndexCompletion);
    }

    /**
     * @param id
     * @param traces
     * @param epoch
     * @param waitForIndexCompletion
     */
    public TmfFilteredExperiment(String id, ITmfTrace[] traces, Object filter, TmfTimestamp startTime, TmfTimestamp endTime, TmfTimestamp epoch, int indexPageSize, boolean waitForIndexCompletion) {
    	super(id, traces, epoch, indexPageSize, waitForIndexCompletion);
    	fFilter = filter;
    	fStartTime = startTime;
    	fEndTime = endTime;

//    	TmfSignalManager.addListener( this );
//    	
		updateNbEvents();
		updateTimeRange();
//		indexExperiment();
    }

    // ========================================================================
    // Accessors
    // ========================================================================

    public int getNbEvents() {
    	int lastCP = fCheckpoints.size() - 1;
    	if( lastCP >= 0 )
    		return (int) fCheckpoints.get( lastCP ).getIndex();
    	return 0;
    }

    /**
     * Returns the index of the first event with the requested timestamp.
     * If none, returns the index of the next event (if any).
     *  
     * @param ts
     * @return
     */
    public long getIndex(TmfTimestamp ts) {

    	return getIndex( ts, null );
		
    }
    
    /**
     * Returns the index of the first event with the requested timestamp.
     * If none, returns the index of the next event (if any).
     * The timestamp of the event at the returned index is put into array
     * element [0] of the tsOut array.
     * 
     * @param ts
     * @param tsOut
     * @return
     */
//    public long getIndex1(final TmfTimestamp ts, TmfTimestamp[] tsOut)
//    {
//    	long time = System.nanoTime();
//    	
//		// Compute the index page and corresponding index
//		int page = Collections.binarySearch( fCheckpoints, new TmfExperimentCheckpoint( 0, ts, null ), new SearchByTimestamp() );
//		
//		// When an exact match on a checkpoint was found, go to previous page 
//		// because we must find and return the first element.
//		while( page >= 0 && fCheckpoints.elementAt( page ).getTimestamp().compareTo( ts, false ) == 0 )
//			page--;
//		
//        // In the very likely case that the checkpoint was not found, bsearch
//        // returns its negated would-be location (not an offset...). From that
//        // index, we can then position the stream and get the event.
//		if( page < 0 ) {
//			page = Math.max( 0, -( page + 2 ) );
//		}
//		
//		long current;
//
//		// Retrieve the checkpoint and set the contexts (make copies)
//		ITmfTrace[] traces = new ITmfTrace[0];	// The set of traces
//		traces = fTraces.toArray( traces );
//		TmfTraceContext[] contexts = new TmfTraceContext[traces.length];
//		if( page >= fCheckpoints.size()) {
//			// If the page entry doesn't exist (e.g. indexing not completed),
//			// set contexts at the the last entry (if it exists)
//			page = fCheckpoints.size() - 1;
//		}
//		if( page >= 0 ) {
//			current = fCheckpoints.elementAt( page ).getIndex();
//			TmfTraceContext[] saveContexts = fCheckpoints.elementAt(page).getContexts();
//			for (int i = 0; i < contexts.length; i++) {
//				contexts[i] = new TmfTraceContext(saveContexts[i]);
//			}
//		} else {
//			// Index is empty... position traces at their beginning
//			current = 0L;
//			for (int i = 0; i < contexts.length; i++) {
//				contexts[i] = new TmfTraceContext(traces[ i ].seekLocation(null));
//				if( fFilter != null ) 
//					traces[ i ].skipEvents( contexts[i], fIndexPageSize, null, fFilter, null );
//			}
//		}
//		long currentCP = current;
//
//		// Position the traces at the requested index
//		TmfEvent event;
//		for( EventIterator i = getEventIterator( traces, contexts ); 
//			( event = i.getNextEvent() ) != null && event.getTimestamp().compareTo( ts, false ) < 0; 
//			current++ );
//		
////		TmfEvent event;
////		for( event = getNextEvent( traces, contexts );
////			event != null && event.getTimestamp().compareTo( ts, false ) < 0;
////			event = getNextEvent( traces, contexts ))
////			current++;
//		
//		time = System.nanoTime() - time;
//		System.out.println("getIndex(" + ts + ") = " + current + "  using checkpoint " + page + " (index=" + currentCP + "); elapsed time=" + time/1000000.0 + "ms");
//		if( tsOut != null )
//			tsOut[ 0 ] = event != null ? event.getTimestamp() : TmfTimestamp.BigCrunch;
//		return current;
//		
//    }

    
    public long getIndex(TmfTimestamp ts, TmfTimestamp[] tsOut)
    {
//    	long time = System.nanoTime();

    	GetIndexHelper getIndexHelper = null;
    	synchronized( this )
    	{
        	if( f_getIndexHelper != null && f_getIndexHelper.appliesTo( ts ))
        	{
            	getIndexHelper = f_getIndexHelper;
            	f_getIndexHelper = null;
        	}
    	}
    	if( getIndexHelper == null )
    		getIndexHelper = new GetIndexHelper( ts );

//		long currentCP = getIndexHelper.current;

		getIndexHelper.moveTo( ts );
    	long current = getIndexHelper.current;
		if( tsOut != null )
			tsOut[ 0 ] = getIndexHelper.tsEvent;

//		time = System.nanoTime() - time;
//		System.out.println("getIndex(" + ts + ") = " + current + "  using checkpoint " + getIndexHelper.checkpoint + " (index=" + currentCP + "); tsOut=" + getIndexHelper.tsEvent + "; elapsed time=" + time/1000000.0 + "ms");

    	synchronized( this )
		{
    		f_getIndexHelper = getIndexHelper;
		}
    	return current;
    }

    
    GetIndexHelper f_getIndexHelper;
    
    
    class GetIndexHelper
    {
    	TmfTimestamp tsStart;
    	TmfTimestamp tsEvent;
    	ITmfTrace[] traces;
    	TmfTraceContext[] contexts;
    	int checkpoint;
   		long current;
		Vector<TmfExperimentCheckpoint> checkpoints;
		EventIterator it;

    	GetIndexHelper( final TmfTimestamp ts )
    	{
    		synchronized( fIndexing )
			{
    			if( fIndexing )
					try
					{
						fIndexing.wait();
					}
					catch( InterruptedException e )
					{
					}
	    		checkpoints = fCheckpoints;
			}
    		
    		// Compute the index page and corresponding index
    		checkpoint = Collections.binarySearch( checkpoints, new TmfExperimentCheckpoint( 0, ts, null ), new SearchByTimestamp() );
    		
    		// When an exact match on a checkpoint was found, go to previous page 
    		// because we must find and return the first element.
    		while( checkpoint >= 0 && checkpoints.elementAt( checkpoint ).getTimestamp().compareTo( ts, false ) == 0 )
    			checkpoint--;
    		
            // In the very likely case that the checkpoint was not found, bsearch
            // returns its negated would-be location (not an offset...). From that
            // index, we can then position the stream and get the event.
    		if( checkpoint < 0 ) {
    			checkpoint = Math.max( 0, -( checkpoint + 2 ) );
    		}
    		
    		// Retrieve the checkpoint and set the contexts (make copies)
    		traces = new ITmfTrace[0];	// The set of traces
    		traces = fTraces.toArray( traces );
    		contexts = new TmfTraceContext[traces.length];
    		if( checkpoint >= checkpoints.size()) {
    			// If the page entry doesn't exist (e.g. indexing not completed),
    			// set contexts at the the last entry (if it exists)
    			checkpoint = checkpoints.size() - 1;
    		}
    		if( checkpoint >= 0 ) {
    			TmfExperimentCheckpoint checkPnt = checkpoints.elementAt( checkpoint );
    			current = checkPnt.getIndex();
    			TmfTraceContext[] saveContexts = checkPnt.getContexts();
    			for (int i = 0; i < contexts.length; i++) {
    				contexts[i] = new TmfTraceContext(saveContexts[i]);
    			}
    			tsStart = checkPnt.getTimestamp();
    		} else {
    			// Index is empty... position traces at their beginning
    			current = 0L;
    			for (int i = 0; i < contexts.length; i++) {
    				contexts[i] = new TmfTraceContext(traces[ i ].seekLocation(null));
    				if( fFilter != null ) 
    					traces[ i ].skipEvents( contexts[i], fIndexPageSize, null, fFilter, null );
    			}
    			tsStart = TmfTimestamp.BigBang;
    		}
    		it = getEventIterator( traces, contexts );
    		TmfEvent event = it.getNextEvent();
    		if( event != null )
    			tsEvent = event.getTimestamp();
    		else
    			tsEvent = TmfTimestamp.BigCrunch;
    	}
    	
    	/**
    	 * Move the current index state to the first event exactly at or after the given timestamp.
    	 * @param ts	reference timestamp.
    	 */
    	void moveTo( final TmfTimestamp ts )
    	{
			if( tsEvent.compareTo( ts, false ) < 0 )
			{
	    		// Position the traces at the requested index
	    		TmfEvent event;
				do
				{
					tsStart = tsEvent;
		    		current++;
				} while(( event = it.getNextEvent() ) != null && ( tsEvent = event.getTimestamp()).compareTo( ts, false ) < 0 );
	    		if( event == null )
	    			tsEvent = TmfTimestamp.BigCrunch;
			}
    	}
    	
    	/**
    	 * Returns true if the current index state is between the given timestamp (inclusive) and the next checkpoint (exclusive).
    	 *  
    	 * @param ts	reference timestamp.
    	 * @return
    	 */
    	boolean appliesTo( final TmfTimestamp ts )
    	{
    		if( tsStart.compareTo( ts, false ) <= 0 )
    		{
        		if( checkpoint >= 0 && fCheckpoints.equals( checkpoints ) )
        		{
        			if( checkpoint < checkpoints.size() - 1 )
        			{
	        			TmfExperimentCheckpoint checkPnt = checkpoints.elementAt( checkpoint + 1 );
	        			if( checkPnt.getTimestamp().compareTo( ts, false ) > 0 )
	        				return true;
        			} else
        				return true;	// last checkpoint always applies
        		}
    		}
    		return false;
    	}
    }
    
    
    /**
     * Returns the timestamp of the event at the requested index.
     * If none, returns null.
     *  
     * @param index
     * @return
     */
    public TmfTimestamp getTimestamp(int index) {
    	// TODO: Support filtering
    	Assert.isTrue( false );
        // TODO: Go over all the traces
    	ITmfTrace trace = fTraces.firstElement();
    	TmfTraceContext context = trace.seekEvent(index);
    	return context.getTimestamp();
    }

    // ========================================================================
    // Operators
    // ========================================================================

    /**
     * Update the total number of events
     */
    protected void updateNbEvents() {
    	Vector<TmfExperimentCheckpoint> checkpoints = fCheckpoints;
    	if( checkpoints != null && checkpoints.size() > 0 )
    	{
    		// indexing has generated a checkpoint at first and last event
    		fNbEvents = (int)( checkpoints.get( checkpoints.size() - 1 ).getIndex() - checkpoints.get( 0 ).getIndex() );
    	}
    }

    /**
     * Update the global time range
     */
	protected void updateTimeRange() {
		super.updateTimeRange();
		if( fStartTime != null && fStartTime.compareTo( fTimeRange.getStartTime(), false ) >= 0 )
			fTimeRange = new TmfTimeRange( fStartTime, fTimeRange.getEndTime() );
		if( fEndTime != null && fEndTime.compareTo( fTimeRange.getEndTime(), false ) <= 0 )
			fTimeRange = new TmfTimeRange( fTimeRange.getStartTime(), fEndTime );
    }
	
    // ========================================================================
    // ITmfRequestHandler
    // ========================================================================
	
	static class SearchByTimestamp implements Comparator<TmfExperimentCheckpoint>
	{

		@Override
		public int compare( TmfExperimentCheckpoint o1, TmfExperimentCheckpoint o2 )
		{
			return o1.getTimestamp().compareTo( o2.getTimestamp(), false );
		}
		
	}
	
	/**
	 * Given an experiment event index, position the set of traces so a call
	 * to getNextEvent() will retrieve the corresponding event.
	 * 
	 * @param index
	 * @param traces
	 * @param contexts
	 * @param nextEvents
	 */
	protected synchronized void positionTraces(long index, ITmfTrace[] traces, TmfTraceContext[] contexts) {

		// Compute the index page and corresponding index
		int page = Collections.binarySearch( fCheckpoints, new TmfExperimentCheckpoint( index, null, null ), new SearchByIndex() );
		
        // In the very likely case that the checkpoint was not found, bsearch
        // returns its negated would-be location (not an offset...). From that
        // index, we can then position the stream and get the event.
		if( page < 0 ) {
			page = Math.max( 0, -( page + 2 ) );
		}
		
		long current;

		// Retrieve the checkpoint and set the contexts (make copies)
		TmfTraceContext[] saveContexts = new TmfTraceContext[contexts.length];
		if( page >= fCheckpoints.size()) {
			// If the page entry doesn't exist (e.g. indexing not completed),
			// set contexts at the the last entry (if it exists)
			page = fCheckpoints.size() - 1;
		}
		if( page >= 0 ) {
			current = fCheckpoints.elementAt( page ).getIndex();
			saveContexts = fCheckpoints.elementAt(page).getContexts();
			for (int i = 0; i < contexts.length; i++) {
				contexts[i] = new TmfTraceContext(saveContexts[i]);
			}
		} else {
			// Index is empty... position traces at their beginning
			current = 0;
			for (int i = 0; i < contexts.length; i++) {
				contexts[i] = new TmfTraceContext(traces[i].seekLocation(null));	// TODO: Handle start time, e.g. seekEvent( fStartTime )
				if( fFilter != null ) 
					traces[i].skipEvents( contexts[i], fIndexPageSize, null, fFilter, null );
			}
		}

		// Position the traces at the requested index
		if( current < index )
		{
			// Setup ordered merging of traces
			TreeSet< NextEvent > nextEvents = new TreeSet< NextEvent >( new OrderNextEventByTimestamp());
			for( int i = 0; i < contexts.length; i++ )
			{
				if( contexts[ i ].getTimestamp() != null && contexts[ i ].getTimestamp().compareTo( fEndTime, false ) <= 0 )
					nextEvents.add( new NextEvent( i, traces[ i ], contexts[ i ]) );
			}
			
			// Fetch settings of next earliest event
			NextEvent nextEvent = nextEvents.pollFirst();
			// Search loop
			while( nextEvent != null && current < index )
			{
				TmfEvent event = nextEvent.trace.getNextEvent( nextEvent.context );
				if( fFilter != null ) 
					nextEvent.trace.skipEvents( nextEvent.context, fIndexPageSize, null, fFilter, null );
				if( nextEvent.trace.testEvent( event, fFilter ) )
				{
					if( ++current >= index )
						break;
				}

				// test if trace does not make any further contribution
				if( nextEvent.context.getTimestamp() != null && nextEvent.context.getTimestamp().compareTo( fEndTime, false ) <= 0 )
				{
					// re-insert current and fetch new settings only when more than one trace
					// and the old is not the earliest one (not the next one)
					if( nextEvents.isEmpty() || 
						nextEvents.comparator().compare( nextEvent, nextEvents.first()) < 0 )
					{
						continue;	// nextEvents is fine, continue with next loop iteration 
					}
					nextEvents.add( nextEvent );	// re-insert current settings (ordered !!!)
				}
				nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
			}					
		}
	}

	/**
	 * Scan the next events from all traces and return the next one
	 * in chronological order.
	 * 
	 * @param traces
	 * @param contexts
	 * @return
	 */
	protected TmfEvent getNextEvent(ITmfTrace[] traces, TmfTraceContext[] contexts) {
		// TODO: Consider the time adjustment
		// Setup ordered merging of traces
		TreeSet< NextEvent > nextEvents = new TreeSet< NextEvent >( new OrderNextEventByTimestamp());
		for( int i = 0; i < contexts.length; i++ )
		{
			if( contexts[ i ].getTimestamp() != null && contexts[ i ].getTimestamp().compareTo( fEndTime, false ) <= 0 )
				nextEvents.add( new NextEvent( i, traces[ i ], contexts[ i ]) );
		}
		
		// Fetch settings of next earliest event
		NextEvent nextEvent = nextEvents.pollFirst();
		// Search loop
		while( nextEvent != null )
		{
			TmfEvent event = nextEvent.trace.getNextEvent( nextEvent.context );
			if( fFilter != null ) 
				nextEvent.trace.skipEvents( nextEvent.context, fIndexPageSize, null, fFilter, null );
			if( nextEvent.trace.testEvent( event, fFilter ) )
			{
				return event;
			}

			// test if trace does not make any further contribution
			if( nextEvent.context.getTimestamp() != null && nextEvent.context.getTimestamp().compareTo( fEndTime, false ) <= 0 )
			{
				// re-insert current and fetch new settings only when more than one trace
				// and the old is not the earliest one (not the next one)
				if( nextEvents.isEmpty() || 
					nextEvents.comparator().compare( nextEvent, nextEvents.first()) < 0 )
				{
					continue;	// nextEvents is fine, continue with next loop iteration 
				}
				nextEvents.add( nextEvent );	// re-insert current settings (ordered !!!)
			}
			nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
		}				
		return null;
	}

	@Override
	protected EventIterator getEventIterator( ITmfTrace[] traces, TmfTraceContext[] contexts )
	{
		return new FilteredEventIterator( traces, contexts );
	}
	
	class FilteredEventIterator extends EventIterator
	{

		public FilteredEventIterator( ITmfTrace[] traces, TmfTraceContext[] contexts )
		{
			// Setup ordered merging of traces
			nextEvents = new TreeSet< NextEvent >( new OrderNextEventByTimestamp());
			for( int i = 0; i < contexts.length; i++ )
			{
				if( contexts[ i ].getTimestamp() != null && contexts[ i ].getTimestamp().compareTo( fEndTime, false ) <= 0 )
					nextEvents.add( new NextEvent( i, traces[ i ], contexts[ i ]) );
			}

			nextEvent = null;
		}
		
		@Override
		public TmfEvent getNextEvent()
		{
			if( nextEvent != null )
			{
				if( nextEvent.context.getTimestamp() != null && nextEvent.context.getTimestamp().compareTo( fEndTime, false ) <= 0 )
				{
					// re-insert current and fetch new settings only when more than one trace
					// and the old is not the earliest one (not the next one)
					if( !nextEvents.isEmpty() && 
						nextEvents.comparator().compare( nextEvent, nextEvents.first()) >= 0 )
					{
						nextEvents.add( nextEvent );	// re-insert current settings (ordered !!!)
						nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
					}
				} else
					nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
			} else
				nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
			// Search loop
			while( nextEvent != null )
			{
				TmfEvent event = nextEvent.trace.getNextEvent( nextEvent.context );
				if( fFilter != null ) 
					nextEvent.trace.skipEvents( nextEvent.context, fIndexPageSize, null, fFilter, null );
				if( nextEvent.trace.testEvent( event, fFilter ) )
				{
					return event;
				}

				// test if trace does not make any further contribution
				if( nextEvent.context.getTimestamp() != null && nextEvent.context.getTimestamp().compareTo( fEndTime, false ) <= 0 )
				{
					// re-insert current and fetch new settings only when more than one trace
					// and the old is not the earliest one (not the next one)
					if( nextEvents.isEmpty() || 
						nextEvents.comparator().compare( nextEvent, nextEvents.first()) < 0 )
					{
						continue;	// nextEvents is fine, continue with next loop iteration 
					}
					nextEvents.add( nextEvent );	// re-insert current settings (ordered !!!)
				}
				nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
			}				
			
			return null;
		}
		
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "[TmfFilteredExperiment (" + fExperimentId + ")]";
	}

    // ========================================================================
    // Indexing
    // ========================================================================

	/*
	 * The experiment holds the globally ordered events of its set of traces.
	 * It is expected to provide access to each individual event by index i.e.
	 * it must be possible to request the Nth event of the experiment.
	 * 
	 * The purpose of the index is to keep the information needed to rapidly
	 * restore the traces contexts at regular intervals (every INDEX_PAGE_SIZE
	 * event).
	 */

	/**
	 * indexExperiment
	 * 
	 * Creates the experiment index.
	 */
	@Override
	public void indexExperiment() {	// supports filtering

		synchronized(fIndexing) {
			if (fIndexing) {
				// An indexing job is already running but a new request came
				// in (probably due to a change in the trace set). The index
				// being currently built is therefore already invalid.
				// TODO: Cancel and restart the job
				// TODO: Add support for dynamically adding/removing traces
				return;
			}
			fIndexing = true;
		}

		job = new IndexingJob(fExperimentId);
		job.schedule();
//		try {
//			job.join();
//		} catch (InterruptedException e) {
//			// TODO Auto-generated catch block
//			e.printStackTrace();
//		}
		if (fWaitForIndexCompletion) {
	        ProgressMonitorDialog dialog = new ProgressMonitorDialog(null);
	        try {
				// TODO: Handle cancel!
	            dialog.run(true, true, new IRunnableWithProgress() {
	                public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
	                    monitor.beginTask("Indexing " + fExperimentId, IProgressMonitor.UNKNOWN);
	    				job.join();
	                    monitor.done();
	                }
	            });
	        } catch (InvocationTargetException e) {
	            e.printStackTrace();
	        } catch (InterruptedException e) {
	            e.printStackTrace();
	        }
		}
	}

	private class IndexingJob extends Job {	// supports filtering

		public IndexingJob(String name) {
			super(name);
		}

		/* (non-Javadoc)
		 * @see org.eclipse.core.runtime.jobs.Job#run(org.eclipse.core.runtime.IProgressMonitor)
		 */
		@Override
		protected IStatus run(IProgressMonitor monitor) {

			int nbTraces = fTraces.size();
			
			// Minimal check
			if (nbTraces == 0) {
				synchronized( fIndexing )
				{
					fIndexing.notify();
		            fIndexing = false;
				}

	            broadcastSignal(new TmfExperimentUpdatedSignal(TmfFilteredExperiment.this, TmfFilteredExperiment.this, null));
	            
				return Status.OK_STATUS;
			}

			monitor.beginTask("Indexing " + fExperimentId, IProgressMonitor.UNKNOWN);

			Vector<TmfExperimentCheckpoint> checkpoints = new Vector<TmfExperimentCheckpoint>();
			
			try
			{
				TreeSet< NextEvent > nextEvents = new TreeSet< NextEvent >( new OrderNextEventByTimestamp());
				
				// get start contexts of all traces
				TmfTraceContext savedContexts[] = new TmfTraceContext[ nbTraces ];
				for( int i = 0; i < nbTraces; i++ )
				{
					ITmfTrace trace = fTraces.get( i );
					TmfTraceContext ctx = trace.seekEvent( fStartTime );
					if( fFilter != null ) 
						trace.skipEvents( ctx, Long.MAX_VALUE, null, fFilter, null );
					if( ctx.getTimestamp() != null && ctx.getTimestamp().compareTo( fEndTime, false ) <= 0 )
						nextEvents.add( new NextEvent( i, trace, ctx ) );
					else
						savedContexts[ i ] = new TmfTraceContext( ctx );
				}
	
				long nextIndex = 0;			// current index
				if( !nextEvents.isEmpty())
				{
					// Monitor overhead
					int skipped = fIndexPageSize;	// force first match to be a checkpoint, so we know that from 0 to first there is no matching event
					// Fetch settings of next earliest event
					NextEvent nextEvent = nextEvents.pollFirst();
					// Search loop
					while( nextEvent != null )
					{
						TmfTraceContext oldContext = new TmfTraceContext( nextEvent.context );
						TmfEvent event = nextEvent.trace.getNextEvent( nextEvent.context );
						if( nextEvent.trace.testEvent( event, fFilter ) )
						{
							if( skipped >= fIndexPageSize )
							{
								// update contexts
								savedContexts[ nextEvent.traceIndex ] = oldContext;	// the current trace needs the context *before* calling getNextEvent
								for( NextEvent otherEvent : nextEvents )
									if( savedContexts[ otherEvent.traceIndex ] == null )
										savedContexts[ otherEvent.traceIndex ] = new TmfTraceContext( otherEvent.context );
								checkpoints.add( new TmfExperimentCheckpoint( nextIndex, oldContext.getTimestamp(), savedContexts.clone() ) );
								skipped = 0;
							}
							nextIndex++;
						}
						savedContexts[ nextEvent.traceIndex ] = null;	// the saved context is outdated and needs to be re-copied when next used for checkpointing
						skipped++;
						if( fFilter != null ) 
							skipped += (int)nextEvent.trace.skipEvents( nextEvent.context, Long.MAX_VALUE, null, fFilter, null );

						monitor.worked( 1 );
						if( monitor.isCanceled() )
						{
//							monitor.done();
//							return Status.CANCEL_STATUS;
							break;
						}

						// test if trace does not make any further contribution
						if( nextEvent.context.getTimestamp() != null && nextEvent.context.getTimestamp().compareTo( fEndTime, false ) <= 0 )
						{
							// re-insert current and fetch new settings only when more than one trace
							// and the old is not the earliest one (not the next one)
							if( nextEvents.isEmpty() || 
								nextEvents.comparator().compare( nextEvent, nextEvents.first()) < 0 )
							{
								continue;	// nextEvents is fine, continue with next loop iteration 
							}
							nextEvents.add( nextEvent );	// re-insert current settings (ordered !!!)
						} else
						{
							// trace is at end and will not be checked any further
							savedContexts[ nextEvent.traceIndex ] = new TmfTraceContext( nextEvent.context ); // make sure a non-null context is set as final = last context
						}
						nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
					}					
				}
				// update contexts
				for( NextEvent otherEvent : nextEvents )	// nextEvents shall be empty here, but just in case...
					if( savedContexts[ otherEvent.traceIndex ] == null )
						savedContexts[ otherEvent.traceIndex ] = new TmfTraceContext( otherEvent.context );
				checkpoints.add( new TmfExperimentCheckpoint( nextIndex, TmfTimestamp.BigCrunch, savedContexts.clone() ) );
			}			
			catch( Throwable t)
			{
				t.printStackTrace();
			}
		
            monitor.done();
			
            synchronized( fIndexing )
			{
                fCheckpoints = checkpoints;
                fIndexing.notify();
	            fIndexing = false;
			}
            
            broadcastSignal(new TmfExperimentUpdatedSignal(TmfFilteredExperiment.this, TmfFilteredExperiment.this, null));
            
            return Status.OK_STATUS;
			
		}
    }

//	/**
//	 * Dump the experiment index
//	 */
//	private void dumpIndex() {
//		System.out.println("-----");
//		System.out.println("Index of " + fExperimentId);
//		for (int i = 0; i < fExperimentIndex.size(); i++) {
//        	System.out.println("Entry:" + i);
//        	TmfTraceContext[] contexts = fExperimentIndex.get(i);
//        	int nbEvents = 0;
//        	for (int j = 0; j < contexts.length; j++) {
//        		ITmfTrace trace = fTraces.get(j);
//            	TmfTraceContext context = trace.seekLocation(contexts[j].getLocation());
//            	TmfEvent event = fTraces.get(j).getNextEvent(new TmfTraceContext(context));
//            	nbEvents += contexts[j].getIndex();
//            	System.out.println("  ["  + trace.getName() + "]" 
//            			+  " index: "     + contexts[j].getIndex()
//            			+ ", timestamp: " + contexts[j].getTimestamp()
//            			+ ", event: "     + event.getTimestamp());
//            	assert (contexts[j].getTimestamp().compareTo(event.getTimestamp(), false) == 0);
//        	}
//        	assert ((i+1) * fIndexPageSize == nbEvents);
//	
//        }
//	}

    // ========================================================================
    // Signal handlers
    // ========================================================================

//    @TmfSignalHandler
//    public void experimentSelected(TmfExperimentSelectedSignal signal) {
//		fCurrentExperiment = this;
//    	indexExperiment();
//    }

    @TmfSignalHandler
    public void experimentUpdated(TmfExperimentUpdatedSignal signal) {
//    	indexExperiment();
    }

    @TmfSignalHandler
    public void traceUpdated(TmfTraceUpdatedSignal signal) {
    	// TODO: Incremental index update
    	synchronized(this) {
    		updateNbEvents();
    		updateTimeRange();
    	}
//		broadcastSignal(new TmfExperimentUpdatedSignal(this, this, signal.getTrace()));
    }

    // ========================================================================
    // Filter and search support
    // ========================================================================

	/* (non-Javadoc)
	 * @see org.eclipse.linuxtools.tmf.request.ITmfRequestHandler#processRequest(org.eclipse.linuxtools.tmf.request.TmfDataRequest, boolean)
	 */
	public void processRequest(TmfFilteringDataRequest<TmfEvent> request, boolean waitForCompletion) {

    	// TODO: Support filtering
    	Assert.isTrue( false );
//		long start = System.nanoTime();
		// Process the request
		processDataRequest(request);
		
		// Wait for completion if needed
    	if (waitForCompletion) {
			request.waitForCompletion();
		}
//    	long end = System.nanoTime();
//    	System.out.println("processRequest duration = "+ (double)(end-start) * 1e-9 + " s");
	}

	static class NextEvent2
	{
		final int traceIndex;
		final ITmfTrace trace;
		final TmfTraceContext context;
		TmfTimestamp nextTimestamp;
		int skipped;
		int consumed;
		TmfTimestamp[] skippedTS;

		NextEvent2( int traceIndex, ITmfTrace trace, TmfTraceContext context, TmfTimestamp nextTimestamp, int skipped, TmfTimestamp[] skippedTS )
		{
			this.traceIndex = traceIndex;
			this.trace = trace;
			this.context = context;
			this.nextTimestamp = nextTimestamp;
			this.skipped = skipped;
			this.consumed = 0;
			this.skippedTS = skippedTS;
		}

//		int consumeTo_1( TmfTimestamp ts, boolean inclusive )
//		{
//			// Linear search
//			int comp = inclusive ? 1 : 0;
//			int start = consumed;
//			while( consumed < skipped && skippedTS[ consumed ].compareTo( ts, false ) < comp )
//				++consumed;
//			return consumed - start;
//		}
		
		int consumeTo( TmfTimestamp ts, boolean inclusive )
		{
			// Binary search
			int comp = inclusive ? 1 : 0;
			int first = consumed, last = skipped, test;
			while( first != last )
			{
				test = ( first + last ) / 2;
				if( skippedTS[ test ].compareTo( ts, false ) < comp )
					first = test + 1;
				else
					last = test;
			}
			first = last - consumed;
			consumed = last;
			return first;
		}
	}

	static class OrderNextEventByTimestamp2 implements Comparator< NextEvent2 >
	{
		@Override
		public int compare( NextEvent2 o1, NextEvent2 o2 )
		{
			int r = o1.nextTimestamp.compareTo( o2.nextTimestamp, true );
			if( r == 0 )
				r = o1.traceIndex - o2.traceIndex;
			return r;
		}
	}

	/**
	 * Process a data request
	 * 
	 * @param request
	 */
	protected void processDataRequest(final TmfFilteringDataRequest<TmfEvent> request) {

		// Process the request
		Thread thread = new Thread() {

			@Override
			public void run() {

				// General request parameters
				final TmfTimestamp endTime;
				final long index;
		
				// Initialize request params depending on request type
				if (request.getRange() != null) {
					index = Math.max( getIndex(request.getRange().getStartTime()), request.getIndex());
					endTime = request.getRange().getEndTime();
		    	} else {
		    		index = request.getIndex();
		    		endTime = TmfTimestamp.BigCrunch;
		    	}

				// Key variables
				Object filter = request.getFilter();

				// Initialize the traces array and position the traces
				// at the first requested event
				long endIndex = request.getNbTestedItems();
				if( endIndex < 0 || endIndex > Long.MAX_VALUE - index )
					endIndex = Long.MAX_VALUE;
				else
					endIndex += index;

				ExperimentFiltering p = new ExperimentFiltering()
				{
					// General request information
					int blockSize;
					int nbRequestedEvents;

					// Create the result buffer
					Vector<TmfEvent> events = new Vector<TmfEvent>();
					Vector<Integer> indexes = new Vector<Integer>();
					
					@Override
					protected boolean isCancelled()
					{
						return request.isCancelled();
					}
					

					@Override
					protected boolean addMatchingEvent( TmfEvent event, long experimentIndex, long nbMatchingEvents )
					{
						events.add( event );
						indexes.add( (int)experimentIndex );
						if( nbMatchingEvents % blockSize == 0 )
							pushData( request, events, indexes );
						return ( nbMatchingEvents < nbRequestedEvents );	// false = No more matches requested.
					}

					@Override
					public void run( final Object filter, final long startIndex, final long endIndex, final TmfTimestamp endTime )
					{
						// Extract the general request information
						blockSize = request.getBlockize();
						nbRequestedEvents = request.getNbRequestedEvents();
						if (nbRequestedEvents <= -1) {
							nbRequestedEvents = Integer.MAX_VALUE;
						}
						
						super.run( filter, index, endIndex, endTime );
						
						if (!request.isCancelled() && !request.isFailed()) {
							if( events.size() != 0 )
								pushData( request, events, indexes );
							request.done();
						}
					}
				};
				p.run( filter, index, endIndex, endTime );
				
//				System.out.println("found " + nbEvents + " events.");
			}
		};
		thread.start();
	}
	/**
	 * Format the result data and notify the requester.
	 * Note: after handling, the data is *removed*.
	 * 
	 * @param request
	 * @param events
	 */
	protected void pushData(TmfFilteringDataRequest<TmfEvent> request, Vector<TmfEvent> events, Vector<Integer> indexes) {
		TmfEvent[] result = new TmfEvent[events.size()];
		events.toArray(result);
		request.setData(result);
		int[] resultIndexes = new int[indexes.size()];
		int s = indexes.size();
		for( int i = 0; i < s; i++ )
			resultIndexes[ i ] = indexes.get( i );
		request.setDataIndex( resultIndexes );
		request.handleData();
		events.removeAllElements();
		indexes.removeAllElements();
	}

	public abstract class ExperimentFiltering
	{
		
		public void run( final Object filter, final long startIndex, final long endIndex, final TmfTimestamp endTime ) 
		{

			// Key variables
			ITmfTrace[] traces = getTraces();	// The set of traces
			TmfTraceContext[] contexts;				// The set of trace contexts

			int nbEvents = 0;

			// Initialize the traces array and position the traces
			// at the first requested event
			contexts = new TmfTraceContext[ traces.length ];
			positionTraces(startIndex, traces, contexts);
			long curIndex = startIndex;

			// Get the first match result for each trace
			long lookAhead = Math.min( DEFAULT_INDEX_PAGE_SIZE, endIndex - curIndex );
			TreeSet< NextEvent2 > nextEvents = new TreeSet< NextEvent2 >( new OrderNextEventByTimestamp2());
			for (int i = 0; i < traces.length; i++)
				if( contexts[ i ].getTimestamp() != null )	// not at end of trace?
				{
					TmfTimestamp[] skippedTS = new TmfTimestamp[ DEFAULT_INDEX_PAGE_SIZE ];
					int skipped = (int)traces[ i ].skipEvents( contexts[ i ], lookAhead, null, filter, skippedTS );
					TmfTimestamp nextTimestamp = contexts[ i ].getTimestamp();
					if( nextTimestamp == null && skipped != 0 )
						nextTimestamp = TmfTimestamp.BigCrunch;
					if( nextTimestamp != null )
						nextEvents.add(  new NextEvent2( i, traces[ i ], contexts[ i ], nextTimestamp, skipped, skippedTS ));
				}

			if( !nextEvents.isEmpty())
			{
				// Fetch settings of next earliest event
				NextEvent2 nextEvent = nextEvents.pollFirst();
				// Search loop
				while( !isCancelled() && nextEvent != null )
				{
					TmfTimestamp skipTo = nextEvent.context.getTimestamp();
					if( skipTo == null )
					{
						// There is no next match, all traces still in
						// the list do not contribute matching events.
						break;
					}
					if( skipTo.compareTo( endTime, false ) >= 0 )
					{
						// Next match exceeds end time.
						break;
					}
					// Advance all other traces to the point of the earliest
					// event
					for( NextEvent2 otherEvent : nextEvents )
						curIndex += otherEvent.consumeTo( skipTo, otherEvent.traceIndex < nextEvent.traceIndex );
					curIndex += nextEvent.skipped - nextEvent.consumed;
					if( curIndex >= endIndex )
					{
						// We skip block of events, so we might exceed
						// the end index!
						break;
					}
					// The earliest event must not be a matching event, it
					// can also happen that there was no matching event within
					// the look-ahead range.
					lookAhead = Math.min( DEFAULT_INDEX_PAGE_SIZE, endIndex - curIndex ); // new look-ahead range
//					long skipped = nextEvent.trace.skipEvents( nextEvent.context, lookAhead, null, filter );
//					if( skipped == 0 )
//					{
						TmfEvent event = nextEvent.trace.getNextEvent( nextEvent.context );
						if( nextEvent.trace.testEvent( event, filter ) )
						{
							++nbEvents;
							if( !addMatchingEvent( event, curIndex, nbEvents ))
								break; // No more matches requested.
						}
						curIndex++;
						nextEvent.skipped = (int)nextEvent.trace.skipEvents( nextEvent.context, lookAhead, null, filter, nextEvent.skippedTS );
						nextEvent.consumed = 0;
//					} else
//					{
//						nextEvent.skipped = skipped;
//					}

					// test if trace does not make any further contribution
					nextEvent.nextTimestamp = nextEvent.context.getTimestamp();
					if( nextEvent.nextTimestamp == null && nextEvent.skipped > 0 )
						nextEvent.nextTimestamp = TmfTimestamp.BigCrunch;
					if( nextEvent.nextTimestamp != null )
					{
						// re-insert current and fetch new settings only when more than one trace
						// and the old is not the earliest one (not the next one)
						if( nextEvents.isEmpty() || 
							nextEvents.comparator().compare( nextEvent, nextEvents.first()) < 0 )
						{
							continue;	// nextEvents is fine, continue with next loop iteration 
						}
						nextEvents.add( nextEvent );	// re-insert current settings (ordered !!!)
					}
					nextEvent = nextEvents.pollFirst();	// fetch new settings of next earliest event
				}
			}
		}
		
		protected boolean isCancelled()
		{
			return false;
		}
		
		abstract protected boolean addMatchingEvent( TmfEvent event, long experimentIndex, long nbMatchingEvents );
		
	}
	
}